/*  pskcodec.c
 *
 *  RTTY coder/decoder functions of xritty application
 */

/*
 *  xritty: An application to transmit and receive
 *  RTTY signals using a computer's sound card
 *
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "codec.h"
#include "shared.h"

/* Index to transmit buffer */
static int xmit_buff_idx;

/* cos lookup wavetable */
static short *cosine_table = NULL;

/* The phase angle Omega for mark
 * and space signal generation */
static double omega = 0.0;

/*------------------------------------------------------------------------*/

/*  Decode_RTTY_Character()
 *
 *  Decodes a baudot-coded character from the
 *  sequence of mark/space signals in RTTY
 *  signal, producing an ASCII equivalent
 */

  gboolean
Decode_RTTY_Character( int *dec_char )
{
  static int ita2_code[ NUM_ITA2_CHARS ] = ITA2_CODE;

  /* Mark and space signal level */
  unsigned char mark, space;

  /* Circular buffers for mark
   * space and average signal */
  static int
	mark_lev[ DECODER_BUFF_SIZE ],
	space_lev[ DECODER_BUFF_SIZE ],
	signal_ave[ DECODER_BUFF_SIZE ];

  /* Variables used by the natched filter that identifies
   * the RTTY character code which best matches the pattern
   * of signal samples in the mark/space signal buffers */
  int
	char_filter_idx,  /* Buffers index for RTTY character matched filter */
	char_filter_lev,  /* Signal level produced by matched filters */
	char_filter_max,  /* The highest signal level found by the filter */
	char_filter_max_idx,  /* ITA2 character index of highest signal sum */
	char_filter_ita2_idx, /* Buffers index used by matched filters */
	char_filter_sum_idx;  /*Count number of RTTY data bits used by filter */

  /* Variables used by the natched filter that identifies
   * the pattern of signal samples in the (mark - space)
   * buffer which best matches the pattern of the sync train */
  int
	sync_filter_lev,	 /* Sync train level sumed up by matched filter */
	sync_filter_sum_idx, /* Index used by matched filter to sum sync train */
	sync_filter_cnt,	 /* Counts the number of buffer locations used above */
	sync_filter_max = 0, /* Maximum signal level found by sync train filter */
	sync_filter_idx = 0; /* Index to samples buffer for sync train filter */

  int
	look_ahead_idx,
	look_ahead_cnt,
	look_ahead_mark_max,
	look_ahead_space_max;

  static int dsp_diversity = 0;

  /* Total number of signal samples per RTTY character.
   * Four samples of signal level are taken for each
   * element in an RTTY character: (1 start bit,
   * 5 data bits and 1 or 1.5 or 2 stop bits) */
  static int samples_per_stopbit;

  static int
	input_idx    = 0,	/* Index to input signal samples buffers   */
	search_idx   = 0,	/* Index to start of search for sync train */
	startbit_idx = 0,	/* Detected position of sync train in buffers */
	buffer_len   = 0,	/* Length in signal samples of signal buffers */
	samples_cnt  = 0,	/* Count of signal samples input to buffers   */
	last_char    = 0;	/* Saves the the last RTTY character received */

  int sync = 0;


  /* Initialize at first call */
  static double stop_bit = 0.0;
  if( stop_bit != rc_data.stop_bit  )
  {
	stop_bit = rc_data.stop_bit;

	/* Samples per stop bit to look for sync train */
	samples_per_stopbit =
	  (int)( (double)LEVELS_PER_BIT * stop_bit + 0.5 );

	/* Length of samples buffers, for 3 RTTY characters */
	buffer_len = NUM_BUFFERED_CHARS * rc_data.samples_per_char;

	/* Wait for 3 characters to be input to buffers */
	samples_cnt = -( 2 * rc_data.samples_per_char );
  } /* if( stop_bit != rc_data.stop_bit ) */

  /* Take a signal sample over 1/4 of RTTY element */
  Afsk_Detect( &mark, &space );

  /* Display mark and space signals */
  if( isFlagSet(ENABLE_DISPLAY) )
	Display_Signal( mark / SCOPE_SCALE, space / SCOPE_SCALE );

  /* Save samples in ring buffers */
  mark_lev[  input_idx ] = (int)mark;
  space_lev[ input_idx ] = (int)space;

  /* Keep sliding average of RTTY signal samples.
   * Space is considered a negative signal here */
  signal_ave[ input_idx ] *= ( SIGNAL_AVE_WINDOW - 1 );
  signal_ave[ input_idx ] += (int)mark - (int)space - dsp_diversity;
  signal_ave[ input_idx ] /= SIGNAL_AVE_WINDOW;

  /* Increment buffer index and cycle */
  input_idx++;
  if( input_idx >= buffer_len ) input_idx = 0;

  /* For every samples_per_bit total samples read in from
   * the detector, try to decode an RTTY character using
   * the start bit detected previously. Then look for a
   * start bit, to be used with the next samples_per_bit
   * total read in. This to make sure that samples for a
   * complete RTTY character train have been read */
  if( samples_cnt >= rc_data.samples_per_char )
  {
	/* Look for a start bit in the samples train */
	/* Find the highest average signal level summed over the
	 * sync train length. The start bit is negated as its space */
	sync_filter_max = -10000;
	for( sync_filter_cnt = 0;
		sync_filter_cnt < rc_data.samples_per_char;
		sync_filter_cnt++ )
	{
	  /* Setup variables for the sync train filter */
	  sync_filter_lev = 0;
	  sync_filter_idx = search_idx;

		/* Summate signal averages over the length of stop bit */
		for( sync_filter_sum_idx = 0;
			sync_filter_sum_idx < samples_per_stopbit;
			sync_filter_sum_idx++ )
		{
		  if( sync_filter_idx == startbit_idx )
			sync_filter_lev -= 10000;
		  else
			sync_filter_lev += signal_ave[ sync_filter_idx ];
		  sync_filter_idx++;
		  if( sync_filter_idx >= buffer_len ) sync_filter_idx = 0;
		}

		/* Summate signal averages over the length of start bit */
		for( sync_filter_sum_idx = 0;
			sync_filter_sum_idx < LEVELS_PER_BIT;
			sync_filter_sum_idx++ )
		{
		  if( sync_filter_idx == startbit_idx )
			sync_filter_lev -= 10000;
		  else
			sync_filter_lev -= signal_ave[ sync_filter_idx ];
		  sync_filter_idx++;
		  if( sync_filter_idx >= buffer_len ) sync_filter_idx = 0;
		}

		/* Save max summation value and index */
		if( sync_filter_max < sync_filter_lev )
		{
		  sync_filter_max = sync_filter_lev;
		  sync = sync_filter_idx;
		}

		/* Reset search index to next signal sample average */
		search_idx++;
		if( search_idx >= buffer_len ) search_idx = 0;
	} /* sync_filter_cnt = 0; sync_filter_cnt < rc_data.samples_per_char; ... */
	startbit_idx = sync;

	/* Look ahead over the samples of a character to find
	 * the unbalance between mark and space signals due
	 * to selective fading of mark and space frequencies */
	if( isFlagSet(ENABLE_DIVERSITY) )
	{
	  look_ahead_mark_max  = -1000;
	  look_ahead_space_max = -1000;
	  look_ahead_idx = startbit_idx + STARTBIT_CORRECTION;
	  if( look_ahead_idx >= buffer_len )
		look_ahead_idx = 0;

	  for( look_ahead_cnt = 0;
		  look_ahead_cnt < rc_data.samples_per_char;
		  look_ahead_cnt++ )
	  {
		/* Find the max values of mark and space samples */
		if( look_ahead_mark_max < mark_lev[ look_ahead_idx ] )
		  look_ahead_mark_max = mark_lev[ look_ahead_idx ];
		if( look_ahead_space_max < space_lev[ look_ahead_idx ] )
		  look_ahead_space_max = space_lev[ look_ahead_idx ];

		look_ahead_idx++;
		if( look_ahead_idx >= buffer_len )
		  look_ahead_idx = 0;
	  }
	  dsp_diversity =
		( look_ahead_mark_max - look_ahead_space_max ) / 4;
	}
	else
	  dsp_diversity = 0;

	/* Find which matching filter produces a maximum
	 * signal level summation. This is done when the
	 * circular buffers have signal level samples for
	 * a complete RTTY character, after the position
	 * of the start bit */
	char_filter_max = -10000;
	char_filter_max_idx = 0;
	for( char_filter_ita2_idx = 0;
		char_filter_ita2_idx < NUM_BAUDOT_CHARS;
		char_filter_ita2_idx++ )
	{
	  /* Start summation of levels one RTTY bit after start bit.
	   * The start bit detection algorithm puts the start bit
	   * position about 1-2 signal samples before the first data bit
	   * so a correction is added to take us to the first data bit */
	  char_filter_idx = startbit_idx + STARTBIT_CORRECTION;
	  if( char_filter_idx >= buffer_len ) char_filter_idx -= buffer_len;
	  char_filter_lev = 0;
	  for( char_filter_sum_idx = 0;
		  char_filter_sum_idx < DATABITS_PER_CHAR;
		  char_filter_sum_idx++ )
	  {
		if( char_filter_ita2_idx & (1 << char_filter_sum_idx) )
		  char_filter_lev += -dsp_diversity +
			mark_lev[ char_filter_idx ] - space_lev[ char_filter_idx ];
		else
		  char_filter_lev += dsp_diversity +
			space_lev[ char_filter_idx ] - mark_lev[ char_filter_idx ];

		char_filter_idx += LEVELS_PER_BIT;
		if( char_filter_idx >= buffer_len ) char_filter_idx -= buffer_len;
	  }

	  if( char_filter_max <= char_filter_lev )
	  {
		char_filter_max     = char_filter_lev;
		char_filter_max_idx = char_filter_ita2_idx;
	  }

	} /* char_filter_ita2_idx = 0; char_filter_ita2_idx < NUM_BAUDOT_CHARS; ... */

	/* Go to FIGS shift characters */
	if( isFlagClear(LTRS_SHIFT) )
	  char_filter_max_idx += NUM_BAUDOT_CHARS;

	/* Process decoded character */
	*dec_char = ita2_code[char_filter_max_idx];
	switch( *dec_char )
	{
	  case 'v': /* LTRS shift */
		SetFlag( LTRS_SHIFT );
		Set_Shift_Label();
		*dec_char = NO_CHARACTER;
		break;

	  case '^': /* FIGS shift */
		ClearFlag( LTRS_SHIFT );
		Set_Shift_Label();
		*dec_char = NO_CHARACTER;
		break;

	  case 'e': /* ENQ character */
		*dec_char = NO_CHARACTER;
		break;

	  case '>': /* LF */
		if( last_char == LF )
		  *dec_char = NO_CHARACTER;
		else
		  *dec_char = LF;
		break;

	  case '<': /* CR */
		  *dec_char = NO_CHARACTER;
		break;

	  case 'b': /* BEL character */
		*dec_char = NO_CHARACTER;
		break;

	  case '|': /* NUL character */
		*dec_char = NO_CHARACTER;
		break;

	  case ' ': /* Unshift on space */
		if( isFlagSet(ENABLE_USOS) )
		{
		  SetFlag( LTRS_SHIFT );
		  Set_Shift_Label();
		}
		break;
	} /* switch( *dec_char ) */
	last_char = *dec_char;

	/* Display sync train status */
	sync = sync_filter_max / SYNCHRO_SCALE;
	if( isFlagClear(ENABLE_DISPLAY) )
	  Display_Synchro( startbit_idx, sync );

	/* Reset samples input counter */
	samples_cnt = 0;

  } /* samples_cnt >= rc_data.samples_per_char */
  else
	*dec_char = NO_CHARACTER;

  /* Keep count of element samples stored */
  samples_cnt++;

  /* Don't print if signal level is below squelch */
  if( (double)sync < rc_data.sqlch_thr )
	*dec_char = NO_CHARACTER;

  return( TRUE );
} /* Decode_RTTY_Character() */

/*------------------------------------------------------------------------*/

/* Cosine_Rising_Edge()
 *
 * Fills the Xmit buffer with a raised cosine rising edge
 */
  void
Cosine_Rising_Edge( int which )
{
  /* Index to cosine lookup table */
  int cosine_idx;
  for( cosine_idx = 0; cosine_idx < rc_data.rtty_bit; cosine_idx++ )
  {
	xmit_buffer[xmit_buff_idx] =
	  ( cosine_table[cosine_idx] * Isin_Omega(which) ) >> 15;
	xmit_buff_idx += rc_data.num_chn;
  }

  xmit_buff_len += rc_data.rtty_bit;
} /* Cosine_Rising_Edge() */

/*------------------------------------------------------------------------*/

/* Cosine_Falling_Edge()
 *
 * Fills the Xmit buffer with a raised cosine falling edge
 */
  void
Cosine_Falling_Edge( int which )
{
  /* Index to cosine lookup table */
  int cosine_idx;
  for( cosine_idx = rc_data.rtty_bit-1; cosine_idx >= 0; cosine_idx-- )
  {
	xmit_buffer[xmit_buff_idx] =
	  ( cosine_table[cosine_idx] * Isin_Omega(which) ) >> 15;
	xmit_buff_idx += rc_data.num_chn;
  }

  xmit_buff_len += rc_data.rtty_bit;
} /* Cosine_Falling_Edge() */

/*------------------------------------------------------------------------*/

/* Steady_Tone()
 *
 * Fills Xmit buffer with steady tone
 */
void
Steady_Tone( int tone_dur, int which )
{
  /* Index to cosine lookup table */
  int tone_idx;
  for( tone_idx = 0; tone_idx < tone_dur; tone_idx++ )
  {
	xmit_buffer[xmit_buff_idx] = Isin_Omega( which );
	xmit_buff_idx += rc_data.num_chn;
  }

  xmit_buff_len += tone_dur;
} /* Steady_Tone() */

/*------------------------------------------------------------------------*/

/* Transmit_Preamble()
 *
 *  Transmits a steady tone and then
 *  diddles as the RTTY pre-amble
 */

  gboolean
Transmit_Preamble( void )
{
  SetFlag( LTRS_SHIFT );
  Set_Shift_Label();

  /* Initialize buffer index
   * according to stereo/mono mode */
  xmit_buff_idx = rc_data.use_chn;
  xmit_buff_len = 0;

  /* Fill buffer with cosine rising edge */
  Cosine_Rising_Edge( MARK );

  /* Fill buffer with the steady tone samples */
  Steady_Tone( rc_data.tone_dur, MARK );

  /* Write buffer to DSP, abort on error */
  if( !DSP_Write(xmit_buffer, xmit_buff_len) )
	return( FALSE );

  /* Initialize buffer index
   * according to stereo/mono mode */
  xmit_buff_idx = rc_data.use_chn;
  xmit_buff_len = 0;

  /* Transmit some diddles (LTRS Shift) */
  if( !Transmit_RTTY_Character('v') ||
	  !Transmit_RTTY_Character('v') )
	return( FALSE );

  return( TRUE );
} /* Transmit_Preamble() */

/*------------------------------------------------------------------------*/

/* Transmit_Postamble()
 *
 *  Transmits a steady tone and then
 *  diddles as the RTTY pre-amble
 */

  gboolean
Transmit_Postamble( void )
{
  /* Initialize buffer index
   * according to stereo/mono mode */
  xmit_buff_idx = rc_data.use_chn;
  xmit_buff_len = 0;

  /* Fill buffer with the steady tone samples */
  Steady_Tone( rc_data.tone_dur, MARK );

  /* Fill buffer with cosine rising edge */
  Cosine_Falling_Edge( MARK );

  /* Write buffer to DSP, abort on error */
  if( !DSP_Write(xmit_buffer, xmit_buff_len) )
	return( FALSE );

  return( TRUE );
} /* Transmit_Postamble() */

/*------------------------------------------------------------------------*/

/*  Transmit_Element()
 *
 *  Transmits an RTTY element (bit)
 */

  gboolean
Transmit_Element( int mark )
{
  int idx;

  /* Send a mark tone */
  switch( mark )
  {
	case MARK:
	  /* Fill buffer with samples at mark tone rate */
	  for( idx = 0; idx < rc_data.rtty_bit; idx++ )
	  {
		xmit_buffer[xmit_buff_idx] = Isin_Omega( MARK );
		xmit_buff_idx += rc_data.num_chn;
	  }
	  break;

	case SPACE:
	  /* Fill buffer with samples at space tone rate */
	  for( idx = 0; idx < rc_data.rtty_bit; idx++ )
	  {
		xmit_buffer[xmit_buff_idx] = Isin_Omega( SPACE );
		xmit_buff_idx += rc_data.num_chn;
	  }
  } /* switch( mark ) */

  xmit_buff_len += rc_data.rtty_bit;
  return( TRUE );
} /* Transmit_Element() */

/*------------------------------------------------------------------------*/

/* Rtty_Character()
 *
 * Prepares an RTTY character in the Xmit buffer
 */
  gboolean
Rtty_Character( int index )
{
  int idx;

  /* Add a start bit */
  Transmit_Element( SPACE );

  /* Send the 5 RTTY character bits */
  for( idx = 0; idx < 5; idx++ )
  {
	if( index & 0x01 )
	  Transmit_Element( MARK );
	else
	  Transmit_Element( SPACE );
	index >>= 1;
  }

  /* Add a stop bit */
  Steady_Tone( rc_data.stop_bit_len, MARK );

  /* Write samples buffer to DSP */
  if( !DSP_Write(xmit_buffer, xmit_buff_len) )
	return( FALSE );

  /* Initialize buffer idx according to stereo/mono mode */
  xmit_buff_idx = rc_data.use_chn;
  xmit_buff_len = 0;

  return( TRUE );
} /* Rtty_Character() */

/*------------------------------------------------------------------------*/

  gboolean
Transmit_RTTY_Character( int ascii_chr )
{
  int ita2_code[ NUM_ITA2_CHARS ] = ITA2_CODE;
  int ita2_idx;

  /* If character is idle (diddle) character */
  if( ascii_chr == 'v' )
  {
	/* Transmit Morse code CW ID */
	if( isFlagSet(TRANSMIT_MORSE_ID) )
	{
	  char mesg[MESG_SIZE];

	  /* Transmit postamble */
	  Transmit_Postamble();

	  /* Send Morse code ID */
	  snprintf( mesg, sizeof(mesg),
		  "  DE %s  TU", rc_data.call );
	  Morse_Transmit( mesg );
	  ClearFlag( TRANSMIT_MORSE_ID );

	  /* Transmit preamble */
	  Transmit_Preamble();

	  return( TRUE );
	} /* if( isFlagSet(TRANSMIT_MORSE_ID) ) */
  } /* if( ascii_chr == '^' ) */

  /* CR and LF must be converted to ita2 code */
  if( ascii_chr == CR ) ascii_chr = '<';
  if( ascii_chr == LF ) ascii_chr = '>';

  /* Find the ascii character in the ITA2 array */
  for( ita2_idx = 0; ita2_idx < NUM_ITA2_CHARS; ita2_idx++ )
	if( ita2_code[ ita2_idx ] == ascii_chr )
	  break;

  /* Ignore non-ITA2 characters */
  if( ita2_idx >= NUM_ITA2_CHARS )
	return( TRUE );

  /* Do not toggle LTRS/FIGS Shift for common
   * characters (NUL, LF, CR, FIGS, LTRS) */
  if( (ascii_chr != 0x00) &&
	  (ascii_chr != '>' ) &&
	  (ascii_chr != '<' ) &&
	  (ascii_chr != ' ' ) &&
	  (ascii_chr != '^' ) &&
	  (ascii_chr != 'v' ) )
  {
	/* Reduce the ita2 index to < 32 and set FIGS shift */
	if( ita2_idx >= NUM_BAUDOT_CHARS )
	{
	  ita2_idx -= NUM_BAUDOT_CHARS;

	  /* Set FIGS shift if needed */
	  if( isFlagSet(LTRS_SHIFT) )
	  {
		ClearFlag( LTRS_SHIFT );
		Set_Shift_Label();
		if( !Rtty_Character(FIGS_SHIFT_IDX) ||
			!Rtty_Character(FIGS_SHIFT_IDX) )
		  return( FALSE );
	  }
	} /* if( ita2_idx >= NUM_BAUDOT_CHARS ) */
	else
	{
	  /* Set LTRS shift if needed */
	  if( isFlagClear(LTRS_SHIFT) )
	  {
		SetFlag( LTRS_SHIFT );
		Set_Shift_Label();
		if( !Rtty_Character(LTRS_SHIFT_IDX) ||
			!Rtty_Character(LTRS_SHIFT_IDX) )
		  return( FALSE );
	  }
	} /* if( ita2_idx >= NUM_BAUDOT_CHARS ) */
  } /* if( (ascii_chr != 0x00) && ... */

  /* Prepare RTTY character */
  Rtty_Character( ita2_idx );

  return( TRUE );
} /* Transmit_RTTY_Character() */

/*------------------------------------------------------------------------*/

/*  Morse_Transmit()
 *
 *  Trasnsmits a character string in Morse code
 */

  gboolean
Morse_Transmit( char *mesg )
{
  int
	mesg_len,   /* String length of message to send  */
	mesg_idx,   /* Index into the message string     */
	morse_chr,  /* Morse code character in hex equiv */
	morse_idx,  /* Index into the Morse Hex code     */
	elem_dur,   /* Duration in samples of Morse elem */
	dash_dur,   /* Duration in samples of Morse dash */
	tone_dur,   /* Duration in samples of trans tone */
	idx;

  /* Tables of Morse and ascii characters */
  int
	ascii_char[] = ASCII_CHAR,
	morse_code[] = MORSE_CODE;

  /* Duration in samples of Morse elem */
  elem_dur = rc_data.dsp_rate / rc_data.morse_wpm;
  dash_dur = 3 * elem_dur;

  /* Convert ascii characters to Morse hex code */
  mesg_len = (int)strlen( mesg );
  for( mesg_idx = 0; mesg_idx < mesg_len; mesg_idx++ )
  {
	/* Find equivalent Hex morse code */
	for( idx = 0; idx < NUM_MORSE_CHARS; idx++ )
	  if( mesg[mesg_idx] == ascii_char[idx] )
		break;
	if( idx < NUM_MORSE_CHARS )
	  morse_chr = morse_code[idx];
	else
	  morse_chr = 1;

	/* Find beginning of Morse code (1st '1') */
	for( idx = 8; idx >= 0; idx-- )
	  if( morse_chr & (1 << idx) )
		break;

	/* Transmit dots (1's) or dashes (0's) */
	morse_idx = idx - 1;
	while( morse_idx >= 0 )
	{
	  /* If element is dot ('1') */
	  if( morse_chr & (1 << morse_idx) )
		/* Tone duration in samples - edges */
		tone_dur = elem_dur - rc_data.rtty_2bit;
	  else /* If element is dash ('0') */
		/* Tone duration in samples is triple */
		tone_dur = dash_dur - rc_data.rtty_2bit;

	  /* Init total samples to be generated */
	  xmit_buff_len = 0;

	  /* Initialize buffer idx according to stereo/mono mode */
	  xmit_buff_idx = rc_data.use_chn;

	  /* Fill buffer with cosine shaped rising edge */
	  Cosine_Rising_Edge( TONE );

	  /* Fill buffer with the steady tone samples */
	  Steady_Tone( tone_dur, TONE );

	  /* Fill buffer with cosine shaped falling edge */
	  Cosine_Falling_Edge( TONE );

	  /* Write samples buffer to DSP, abort on error */
	  if( !DSP_Write(xmit_buffer, xmit_buff_len) )
	  {
		free_ptr( (void **)&xmit_buffer );
		return( FALSE );
	  }

	  /* Initialize buffer idx according to stereo/mono mode */
	  xmit_buff_idx = rc_data.use_chn;

	  /* Fill buffer with 0 level for space */
	  /* If inter-word space add 3 extra space elements*/
	  if( morse_chr == 0x01 )
	  {
		for( idx = 0; idx < dash_dur; idx++ )
		{
		  xmit_buffer[xmit_buff_idx] = 0;
		  xmit_buff_idx += rc_data.num_chn;
		}
		xmit_buff_len = dash_dur;

		/*** Writebuffer to DSP, abort on error ***/
		if( !DSP_Write(xmit_buffer, xmit_buff_len) )
		{
		  free_ptr( (void **)&xmit_buffer );
		  return( FALSE );
		}
	  }/*  if( morse_chr == 0x01 ) */

	  /* Add inter-element space */
	  for( idx = 0; idx < elem_dur; idx++ )
	  {
		xmit_buffer[xmit_buff_idx] = 0;
		xmit_buff_idx += rc_data.num_chn;
	  }
	  xmit_buff_len = elem_dur;

	  /*** Write buffer to DSP, abort on error ***/
	  if( !DSP_Write(xmit_buffer, xmit_buff_len) )
	  {
		free_ptr( (void **)&xmit_buffer );
		return( FALSE );
	  }

	  morse_idx--;
	} /* while( morse_idx >= 0 ) */

	/* Send inter-character space */
	/* Initialize buffer according to stereo/mono mode */
	xmit_buff_idx = rc_data.use_chn;
	xmit_buff_len = 3 * elem_dur;
	for( idx = 0; idx < xmit_buff_len; idx++ )
	{
	  xmit_buffer[xmit_buff_idx] = 0;
	  xmit_buff_idx += rc_data.num_chn;
	}

	/*** Write samples buffer to DSP ***/
	if( !DSP_Write(xmit_buffer, xmit_buff_len) )
	{
	  free_ptr( (void **)&xmit_buffer );
	  return( FALSE );
	}

  } /* for( mesg_idx = 0; mesg_idx < mesg_len; mesg_idx++ ) */

  return( TRUE );
} /* Morse_Transmit() */

/*------------------------------------------------------------------------*/

/* Make_Cosine_Table()
 *
 * Builds a raised cosine wavetable for tone generation
 */
  gboolean
Make_Cosine_Table( void )
{
  int idx;
  double step = M_PI_2 / (double)rc_data.rtty_bit;

   if( !mem_alloc((void **)&cosine_table,
	   (size_t)rc_data.rtty_bit * sizeof(short)) )
	return( FALSE );

  /* This makes a raised cosine table */
  for( idx = 0; idx < rc_data.rtty_bit; idx++ )
	cosine_table[idx] = (short)( 32767.0 * sin(step * (double)idx) );

  return(TRUE);
} /* Make_Cosine_Table() */

/*------------------------------------------------------------------------*/

/* Isin_Omega()
 *
 * Provides the next value of sin(omega) as a short
 */
  short
Isin_Omega( int which )
{
  double dW = 0.0;
  short isin;

  switch( which )
  {
	case MARK:
	  dW = rc_data.mark_dW;
	  break;
	case SPACE:
	  dW = rc_data.space_dW;
	  break;
	case TONE:
	  dW = rc_data.tone_dW;
	  break;
  }

  isin = (short)( 32767.0 * sin(omega) );
  omega += dW;
  if( omega >= TWOPI ) omega -= TWOPI;

  return( isin );
} /* Isin_Omega() */

/*------------------------------------------------------------------------*/

