/*  display.c
 *
 *  Display/Rendering functions for xritty
 */

/*
 *  xritty: An application to transmit and receive
 *  RTTY signals using the computer's sound card
 *
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "display.h"
#include "shared.h"

/*------------------------------------------------------------------------*/

  int
DFT_Bin_Value( int sum_i, int sum_q, gboolean reset )
{
 /* Value of dft output "bin" */
  static int bin_val = 0;

  /* Maximum value of dft bins */
  static int bin_max = 1000, max = 0;


  /* Calculate sliding window average of max bin value */
  if( reset )
  {
	bin_max = max;
	if( !bin_max ) bin_max = 1;
	max = 0;
  }
  else
  {
	/* Calculate average signal power at each frequency (bin) */
	bin_val  = bin_val * AMPL_AVE_MUL;
	bin_val += sum_i * sum_i + sum_q * sum_q;
	bin_val /= AMPL_AVE_WIN;

	/* Record max bin value */
	if( max < bin_val ) max = bin_val;

	/* Scale bin values to 255 depending on max value */
	int ret = ( 255 * bin_val ) / bin_max;
	if( ret > 255 ) ret = 255;
	return( ret );
  }

  return( 0 );
} /* DFT_Bin_Value() */

/*------------------------------------------------------------------------*/

/* Display_Waterfall()
 *
 * Displays audio spectrum as "waterfall"
 */

  void
Display_Waterfall( void )
{
  int
	idh, idv,  /* Index to hor. and vert. position in warterfall */
	pixel_val, /* Greyscale value of pixel derived from dft o/p  */
	dft_idx,   /* Index to dft output array */
	temp;

  /* Constants needed to draw white lines in waterfall */
  int mark_line, space_line, range;

  /* Pointer to current pixel */
  static guchar *pix;

  /* Constants needed to draw white lines in waterfall */
  range = rc_data.tone_freq * 2;
  mark_line  = (rc_data.mark_freq  * gbl_wfall_width) / range;
  space_line = (rc_data.space_freq * gbl_wfall_width) / range;

  /* Draw a vertical white line in waterfall at detector's freq. */
  pix = gbl_wfall_pixels + gbl_wfall_rowstride + gbl_wfall_n_channels;
  pix += gbl_wfall_n_channels * mark_line;
  pix[0] = pix[1] = pix[2] = 0xff;
  pix = gbl_wfall_pixels + gbl_wfall_rowstride + gbl_wfall_n_channels;
  pix += gbl_wfall_n_channels * space_line;
  pix[0] = pix[1] = pix[2] = 0xff;

  /* Copy each line of waterfall to next one */
  temp = gbl_wfall_height - 2;
  for( idv = temp; idv > 0; idv-- )
  {
	pix = gbl_wfall_pixels + gbl_wfall_rowstride * idv + gbl_wfall_n_channels;

	for( idh = 0; idh < gbl_wfall_width; idh++ )
	{
	  *pix = *( pix - gbl_wfall_rowstride ); pix++;
	  *pix = *( pix - gbl_wfall_rowstride ); pix++;
	  *pix = *( pix - gbl_wfall_rowstride );
	  pix += gbl_wfall_n_channels - 2;
	}
  }

  /* Got to top left of pixbuf */
  pix = gbl_wfall_pixels;

  /* First (DC) dft bin not used */
  gbl_bin_ave[0] = 0;

  /* Do the DFT on input array */
  Idft( DFT_INPUT_SIZE, gbl_wfall_width );
  for( dft_idx = 0; dft_idx < gbl_wfall_width; dft_idx++ )
  {
	/* Calculate power of signal at each freq. ("bin") */
	dft_out_r[dft_idx] /= 1024;
	dft_out_i[dft_idx] /= 1024;
	pixel_val = DFT_Bin_Value(
		dft_out_r[dft_idx], dft_out_i[dft_idx], FALSE );

	/* Calculate average bin values */
	gbl_bin_ave[dft_idx] = pixel_val;

	/* Color code signal strength */
	int n;
	if( pixel_val < 85 )
	{
	  pix[0] = 0;
	  pix[1] = 0;
	  pix[2] = (guchar)(3 + pixel_val * 3);
	}
	else if( pixel_val < 170 )
	{
	  n = pixel_val - 85;
	  pix[0] = 0;
	  pix[1] = (guchar)(3 + pixel_val * 3);
	  pix[2] = (guchar)(255 - n * 3);
	}
	else
	{
	  n = pixel_val - 170;
	  pix[0] = (guchar)pixel_val;
	  pix[1] = (guchar)255;
	  pix[2] = 0;
	}

	pix += gbl_wfall_n_channels;

  } /* for( dft_idx = 0; dft_idx < DFT_SIZE2; dft_idx++ ) */

  /* Reset function */
  DFT_Bin_Value( 0, 0, TRUE );

  /* At last draw waterfall */
  gtk_widget_queue_draw( gbl_scope1 );

} /* Display_Waterfall() */

/*------------------------------------------------------------------------*/

/*  Display_Synchro()
 *
 *  Updates the Synchro display
 */

  void
Display_Synchro( int sync_pos, int sync_lev )
{
  /* Factors needed to draw startbit position pointer */
  static double
	width  = 0.0,
	height = 0.0,
	sync_height,
	sync_width,
	sync_y,
	sync_x,
	sync_r,
	scope_center_x,
	scope_center_y;

  double ind_x, ind_y;
  double w, sin_w, cos_w;

  /* Initialize on first call or change of params */
  if( (height != (double)gbl_wfall_height) ||
	  (width  != (double)gbl_wfall_width) )
  {
	height = (double)gbl_wfall_height;
	width  = (double)gbl_wfall_width;
	sync_height = height - 14.0;
	sync_width  = round( width / 8.0 );
	sync_y = height - 6.0;
	sync_x = round( width * 0.7 );
	scope_center_x = height / 2.0;
	scope_center_y = height / 2.0;
	sync_r = height / 2.0 - 8.0;
  }

  static cairo_t *cr = NULL;

  /* Create a Cairo context */
  if( cr == NULL )
	cr = gdk_cairo_create( gbl_scope2->window );

  /* Fill scope backround black */
  cairo_set_source_rgb( cr, 0.0, 0.0, 0.0 );
  cairo_rectangle( cr,
	  0.0, 0.0,
	  (double)gbl_wfall_width,
	  (double)gbl_wfall_height );
  cairo_fill( cr );

  /* Calculate co-ordinates for sync position plot */
   w = TWOPI * (double)sync_pos /
	 (double)(rc_data.samples_per_char * NUM_BUFFERED_CHARS);
   sin_w = sin( w);
   cos_w = cos( w );
   ind_x = sync_r * sin_w;
   ind_y = sync_r * cos_w;

  /* Draw a green circle for the synchro scope */
  cairo_set_source_rgb( cr, 0.0, 1.0, 0.0 );
  cairo_arc( cr,
	  scope_center_x,
	  scope_center_y,
	  sync_r,
	  0.0, TWOPI );

  /* Draw the start bit position in the circular buffer */
  cairo_move_to( cr, scope_center_x, scope_center_y );
  cairo_line_to( cr,
	  scope_center_x + ind_x ,
	  scope_center_y - ind_y );
  cairo_stroke( cr );

  /* Show the squelch level in red */
  if( sync_lev < rc_data.sqlch_thr )
	cairo_set_source_rgb( cr, 0.8, 0.0, 0.0 );
  else
	cairo_set_source_rgb( cr, 0.0, 0.8, 0.0 );

  /* Create bar graph of sync level */
  if( (double)sync_lev > sync_height )
	sync_lev = (int)sync_height;
  cairo_rectangle( cr,
	  sync_x,
	  sync_y - (double)sync_lev,
	  sync_width,
	  (double)sync_lev );
  cairo_fill( cr );

  /* Create bar graph outline of sync level */
  cairo_set_source_rgb( cr, 0.0, 0.8, 0.0 );
  cairo_rectangle( cr,
	  sync_x,
	  8.0,
	  sync_width,
	  sync_height );
  cairo_stroke( cr );

} /*  Display_Synchro() */

/*------------------------------------------------------------------------*/

/*  Display_Signal()
 *
 *  Updates the signal scope display
 */

  void
Display_Signal( unsigned char plot1, unsigned char plot2 )
{
  /* Points to plot */
  static GdkPoint *points1 = NULL;
  static GdkPoint *points2 = NULL;
  static int points_idx = 0;
  static cairo_t *cr1 = NULL, *cr2 = NULL;
  int height  = gbl_wfall_height;

  /* Initialize on first call */
  if( points1 == NULL )
  {
	if( !mem_alloc((void **)&points1,
		  sizeof(GdkPoint) * (size_t)gbl_wfall_width) )
	  return;
  }
  if( points2 == NULL )
  {
	if( !mem_alloc((void **)&points2,
		  sizeof(GdkPoint) * (size_t)gbl_wfall_width) )
	  return;
  }

  /* Initialize Cairo */
  if( cr1 == NULL )
	cr1 = gdk_cairo_create( gbl_scope1->window );
  if( cr2 == NULL )
	cr2 = gdk_cairo_create( gbl_scope2->window );

  /* Save values to be plotted */
  if( plot1 > height ) plot1 = (unsigned char)height;
  points1[points_idx].y = height - (gint)plot1 - 1;
  points1[points_idx].x = points_idx;

  if( plot2 > height ) plot2 = (unsigned char)height;
  points2[points_idx].y = height - (gint)plot2 - 1;
  points2[points_idx].x = points_idx;

  /* Recycle buffer idx when full and plot */
  if( ++points_idx == gbl_wfall_width )
  {
	int idx;

	/* Draw scope backgrounds */
	cairo_set_source_rgb( cr1, 0.0, 0.3, 0.0 );
	cairo_rectangle(
		cr1, 0.0, 0.0,
		(double)gbl_wfall_width,
		(double)gbl_wfall_height );
	cairo_fill( cr1 );
	cairo_set_source_rgb( cr2, 0.0, 0.3, 0.0 );
	cairo_rectangle(
		cr2, 0.0, 0.0,
		(double)gbl_wfall_width,
		(double)gbl_wfall_height );
	cairo_fill( cr2 );

	/* Plot signal graph 1 */
	cairo_set_source_rgb( cr1, 0.0, 1.0, 0.0 );

	cairo_move_to( cr1,
		(double)points1[0].x,
		(double)points1[0].y );
	for( idx = 1; idx < points_idx; idx++ )
	  cairo_line_to( cr1,
		  (double)points1[idx].x,
		  (double)points1[idx].y );

	/* Stroke paths */
	cairo_stroke( cr1 );

	/* Plot signal graph 2 */
	cairo_set_source_rgb( cr2, 0.0, 1.0, 0.0 );

	cairo_move_to( cr2,
		(double)points2[0].x,
		(double)points2[0].y );
	for( idx = 1; idx < points_idx; idx++ )
	  cairo_line_to( cr2,
		  (double)points2[idx].x,
		  (double)points2[idx].y );

	/* Stroke paths */
	cairo_stroke( cr2 );

	points_idx = 0;
  } /* if( ++points_idx == wfall_pixbuf.width ) */

} /* Display_Signal( void ) */

/*------------------------------------------------------------------------*/

/*  Set_TxRx_Labels()
 *
 *  Sets up the labels in the Tx/Rx frames
 */

  void
Set_TxRx_Labels( void )
{
  if( isFlagSet(TRANSMIT_KEYBD) ||
	  isFlagSet(TRANSMIT_MACRO) )
	gtk_label_set_markup(
		GTK_LABEL(lookup_widget(main_window, "xmit_status")),
		XMIT_RTTY );
  else gtk_label_set_markup(
	  GTK_LABEL(lookup_widget(main_window, "xmit_status")),
	  XMIT_OFF );

  if( isFlagSet(RECEIVE_MODE) )
	gtk_label_set_markup(
		GTK_LABEL(lookup_widget(main_window, "rcve_status")),
		RECV_RTTY );
  else gtk_label_set_markup(
	  GTK_LABEL(lookup_widget(main_window, "rcve_status")),
	  RECV_OFF );

  /* Close cat if tx/rx off */
  if( isFlagClear(RECEIVE_MODE) &&
	  isFlagClear(TRANSMIT_MACRO) &&
	  isFlagClear(TRANSMIT_KEYBD) )
	Close_Tcvr_Serial();

} /* Set_TxRx_Labels() */

/*------------------------------------------------------------------------*/

/*  Clear_Record_Fields()
 *
 *  Clears all QSO Record fields
 */

  void
Clear_Record_Fields( gboolean all )
{
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "callsign")), "" );
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "rst_in")), "" );
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "rst_out")), "" );
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "op_name")), "" );
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "qth_name")), "" );
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "locator")), "" );
  if( all )
	gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "band")), "" );

  qso_record.dx_call[0] = '\0';
  qso_record.my_rst[0]  = '\0';
  qso_record.dx_rst[0]  = '\0';
  qso_record.dx_qth[0]  = '\0';
  qso_record.dx_loc[0]  = '\0';
  if( all )
	qso_record.freq[0]  = '\0';
  Strlcpy( qso_record.dx_name, "OP", sizeof(qso_record.dx_name) );

  ClearFlag( SAVE_RECORD );
} /*  Clear_Record_Fields() */

/*------------------------------------------------------------------------*/

/*  Print_Character()
 *
 *  Prints a character to a text view port
 */

void
Print_Character(
	int printchr,
	GtkTextBuffer *text_buffer,
	GtkWidget     *scrolledwindow,
	GtkTextIter   *iter )
{
  GtkAdjustment *adjustment;
  char char2text[2];

  /* Delete a char on backspace */
  if( printchr == GDK_BackSpace )
  {
	GtkTextIter start;

	gtk_text_buffer_get_iter_at_offset(
		text_buffer, &start,
		gtk_text_buffer_get_char_count
		(text_buffer) -1 );
	gtk_text_buffer_delete( text_buffer, &start, iter );
	return;
  }

  /* Convert char to "text" */
  char2text[0] = (char) printchr;
  char2text[1] = '\0';

  /* Print character */
  gtk_text_buffer_insert( text_buffer, iter, char2text, -1 );

  /* Record to QSO record file */
  if( isFlagSet(RECORD_QSO)    &&
	  (isFlagSet(RECEIVE_MODE) ||
	   isFlagSet(TRANSMIT_MODE)) )
	fprintf( qso_record_fp, "%c", printchr );

  /* Scroll Text View to bottom on LF */
  if( (printchr == LF) || (printchr == GDK_Return) )
  {
	adjustment = gtk_scrolled_window_get_vadjustment
	  ( GTK_SCROLLED_WINDOW(scrolledwindow) );
	gtk_adjustment_set_value( adjustment, adjustment->upper );

	if( isFlagSet(RECEIVE_MODE) )
	  fprintf( qso_record_fp, "Rx-> " );
	else if( isFlagSet(TRANSMIT_MODE) )
	  fprintf( qso_record_fp, "Tx-> " );
  }

} /* Print_Character() */

/*------------------------------------------------------------------------*/

/* Set_Shift_Label()
 *
 * Sets LTRS or FIGS to the Shift Status label
 */

  void
Set_Shift_Label( void )
{
  GtkLabel *lbl =
	GTK_LABEL( lookup_widget(main_window, "shift_label") );

  if( isFlagSet(LTRS_SHIFT) )
	gtk_label_set_markup( lbl, LTRS );
  else
	gtk_label_set_markup( lbl, FIGS );

} /* Set_Shift_Label() */

/*------------------------------------------------------------------------*/

